import torch
import matplotlib.pyplot as plt
import numpy as np
import itertools

def _print_rules(instance, mean_rule_activation=False):
        '''Returns a list with the rules of the model in str format.
        
        Args:
            mean_rule_activation: bool to return mean rule activation.
            
        Returns:
            rules:                list of str representing the rules of the system.
            mean:                 numpy array (R) with normalized mean rule activation.
        '''
        
        combinations = list(itertools.product(*[range(i) for i in instance.input_n_sets]))
        rules = []

        for i, combination in enumerate(combinations):
            rule = 'IF '
            for var_index, set_index in enumerate(combination):
                if var_index != len(combination) - 1:
                    rule += f'{instance.input_var_names[var_index]} IS {instance.input_mf_names[var_index][set_index]} AND '
                else:
                    rule += f'{instance.input_var_names[var_index]} IS {instance.input_mf_names[var_index][set_index]}, '

            rule += f'THEN {instance.output_var_names} IS f{i}'
            rules.append(rule)
        
        if mean_rule_activation:
            mean = torch.mean(torch.stack(instance.antecedents.mean_rule_activation, dim=0), dim=0)
            mean /= torch.sum(mean)
            mean = mean.numpy()
            
            return rules, mean
            
        return rules

def _plot_var(instance, var_name, file_name=False):
        '''Plots the membership functions for a certain variable of the model.

        Args:
            var_name:  str with the name of the variable, written in the same way as given in dict variables.
            file_name: str with the name of the file to be saved, if desired.
        '''

        var_index = instance.input_var_names.index(var_name)
        n_sets = instance.input_n_sets[var_index]
        uod = torch.linspace(*instance.input_uod[var_index], 1000)
        mf_names = instance.input_mf_names[var_index]
        mf_function = instance.memberships[var_index]

        memberships = mf_function(uod)

        for i in range(n_sets):
            plt.plot(uod.numpy(), memberships[:, i].detach().numpy(), label=f'{mf_names[i]}')

        plt.title(f'Membership Functions for Variable {var_name}')
        plt.xlabel('Universe of Discourse')
        plt.ylabel('Membership')
        plt.legend()

        if isinstance(file_name, str):
            plt.savefig(f'{file_name}.png', bbox_inches='tight')

        plt.show()

def _plot_rules(
        instance, 
        var_names, 
        n_points=1000, 
        thr=0.8, 
        levels=10, 
        cmap='viridis', 
        alpha=0.3,
        x_data=None,
        y_data=None,
        file_name=None,
    ):
        '''Plot the projection of the fuzzy rules in a two variable space.
        
        Args:
            var_names: list/tuple with the variables names.
            n_points:  int with 
            thr:       float between 0 and 1 with the threshold value of the fuzzy rules activation.
            levels:    same as matplotlib.pyplot.
            cmap:      same as matplotlib.pyplot.
            alpha:     same as matplotlib.pyplot.
            x_data:    data for scatter plot.
            y_data:    data for scatter plot.
            file_name: str with the name of the file to be saved, if desired.
        '''
        
        if len(var_names) != 2:
            raise ValueError('Exactly two variable names should be provided.')
        
        var_index = instance.input_var_names.index(var_names[0])
        n_sets0 = instance.input_n_sets[var_index]
        uod0 = torch.linspace(*instance.input_uod[var_index], n_points)
        mf_names0 = instance.input_mf_names[var_index]
        mf_function0 = instance.memberships[var_index]

        memberships0 = mf_function0(uod0)
        
        var_index = instance.input_var_names.index(var_names[1])
        n_sets1 = instance.input_n_sets[var_index]
        uod1 = torch.linspace(*instance.input_uod[var_index], n_points)
        mf_names1 = instance.input_mf_names[var_index]
        mf_function1 = instance.memberships[var_index]

        memberships1 = mf_function1(uod1)
        
        x = np.linspace(uod0[0], uod0[-1], n_points)
        y = np.linspace(uod1[0], uod1[-1], n_points)
        X, Y = np.meshgrid(x, y)
        
        fig = plt.figure(figsize=(7, 7))
        gs = fig.add_gridspec(2, 2, width_ratios=[5, 1.5], height_ratios=[1.5, 5], wspace=0.08, hspace=0.08)

        ax_main = fig.add_subplot(gs[1, 0])  
        ax_top = fig.add_subplot(gs[0, 0])  
        ax_right = fig.add_subplot(gs[1, 1])  
        
        for i in range(n_sets0):
            for j in range(n_sets1):
                mu0 = memberships0[:, i].detach()  
                mu1 = memberships1[:, j].detach()  

                mf0 = mu1.unsqueeze(1).repeat(1, n_points) 
                mf1 = mu0.unsqueeze(0).repeat(n_points, 1)
                rule = torch.stack([mf0, mf1], dim=0)
                Z = instance.and_operator(rule, dim=0).numpy()
                
                Z[Z < thr] = np.nan
                ax_main.contourf(X, Y, Z, levels=levels, cmap=cmap, alpha=alpha)
                
        ax_main.set_xlabel(var_names[0])
        ax_main.set_ylabel(var_names[1])
        
        cbar_ax = fig.add_axes([ax_main.get_position().x0 - 0.16, 
                        ax_main.get_position().y0,  
                        0.02,  
                        ax_main.get_position().height])  

        cbar = fig.colorbar(ax_main.contourf(X, Y, Z, levels=levels, cmap=cmap, alpha=alpha), cax=cbar_ax, orientation="vertical")

        cbar.ax.yaxis.set_ticks_position("left")
        cbar.ax.yaxis.set_label_position("left")
        
        for i in range(n_sets0):
            ax_top.plot(uod0, memberships0[:, i].detach().numpy(), label=f'{mf_names0[i]}', lw=2)
        
        ax_top.legend()
        ax_top.set_xticks([])
        ax_top.set_yticks([])

        for i in range(n_sets1):
            ax_right.plot(memberships1[:, i].detach().numpy(), uod1, label=f'{mf_names1[i]}', lw=2)

        ax_right.legend()
        ax_right.set_xticks([])
        ax_right.set_yticks([])
        
        ax_top.set_xlim(ax_main.get_xlim())
        ax_right.set_ylim(ax_main.get_ylim())

        ax_top.set_frame_on(False)
        ax_right.set_frame_on(False)
        
        if x_data is None:
            pass
        else:
            ax_main.scatter(x_data, y_data, color='red')
        
        if isinstance(file_name, str):
            plt.savefig(f'{file_name}.png', bbox_inches='tight')

        plt.show()



import torch
import torch.nn as nn

class GaussianMF(nn.Module):
    def __init__(self, n_sets, uod):
        '''Gaussian membership function. Receives the input tensor of a variable and outputs the tensor with the membership            
        values.

        Args:
            n_sets:      int for number of fuzzy sets associated to the variable.
            uod:         list/tuple with the universe of discourse of the variable.

        Returns:
            memberships: tensor (N, n) containing the membership values of the inputs.
        '''

        super(GaussianMF, self).__init__()

        self.n_sets = n_sets
        self.uod = uod
        
        step = ((uod[1] - uod[0]) / (n_sets - 1))
        
        self.mu = nn.Parameter(torch.linspace(*uod, n_sets))
        self.sigma = nn.Parameter((step / 2) * torch.ones(n_sets))

    def forward(self, x):
        x = x.reshape(-1, 1).repeat(1, self.n_sets)
        mu = torch.minimum(torch.maximum(self.mu, torch.tensor(self.uod[0])), torch.tensor(self.uod[1]))
        sigma = nn.functional.relu(self.sigma) + 1e-6
        memberships = torch.exp(-((x - mu) ** 2) / (2 * sigma ** 2))
        return memberships
    
class TriangularMF(nn.Module):
    def __init__(self, n_sets, uod):
        '''Triangular membership function. Receives the input tensor of a variable and outputs the tensor with the                      
        membership values.

        Args:
            n_sets: int for number of fuzzy sets associated to the variable.
            uod:    list/tuple with the universe of discourse of the variable.

        Returns:
            memberships: tensor (N, n) containing the membership values of the inputs.
        '''

        super(TriangularMF, self).__init__()

        self.n_sets = n_sets
        self.uod = uod
        
        step = ((uod[1] - uod[0]) / (n_sets - 1))
        
        self.b = nn.Parameter(torch.linspace(*uod, n_sets))
        self.deltaL = nn.Parameter(step * torch.ones(n_sets))
        self.deltaR = nn.Parameter(step * torch.ones(n_sets))

    def forward(self, x, delta=10e-8):
        x = x.reshape(-1, 1).repeat(1, self.n_sets)
        b = torch.minimum(torch.maximum(self.b, torch.tensor(self.uod[0])), torch.tensor(self.uod[1]))
        deltaL = nn.functional.relu(self.deltaL) + 1e-6
        deltaR = nn.functional.relu(self.deltaR) + 1e-6
        left = (x + deltaL - b) / (deltaL + delta)
        right = (deltaR + b - x) / (deltaR + delta)
        memberships = torch.maximum(torch.minimum(left, right), torch.tensor(0.0))
        return memberships

class BellMF(nn.Module):
    def __init__(self, n_sets, uod):
        '''Generalized bell membership function. Receives the input tensor of a variable and outputs the tensor with the 
        membership values.

        Args:
            n_sets:      int for number of fuzzy sets associated to the variable.
            uod:         list/tuple with the universe of discourse of the variable.

        Returns:
            memberships: tensor (N, n) containing the membership values of the inputs.
        '''

        super(BellMF, self).__init__()

        self.n_sets = n_sets
        self.uod = uod
        
        step = ((uod[1] - uod[0]) / (n_sets - 1))
        
        self.c = nn.Parameter(torch.linspace(*uod, n_sets))
        self.a = nn.Parameter((step / 2) * torch.ones(n_sets))
        self.b = nn.Parameter(step * torch.ones(n_sets))

    def forward(self, x):
        x = x.reshape(-1, 1).repeat(1, self.n_sets)
        c = torch.minimum(torch.maximum(self.c, torch.tensor(self.uod[0])), torch.tensor(self.uod[1]))
        a = nn.functional.relu(self.a) + 1e-6
        b = nn.functional.relu(self.b) + 1e-6
        memberships = 1 / (1 + torch.abs((x - c) / (a)) ** (2 * b))
        return memberships
    
class SigmoidMF(nn.Module):
    def __init__(self, n_sets, uod):
        '''Sigmoid membership function. Receives the input tensor of a variable and outputs the tensor with the 
        membership values.

        Args:
            n_sets:      int for number of fuzzy sets associated to the variable.
            uod:         list/tuple with the universe of discourse of the variable.

        Returns:
            memberships: tensor (N, n) containing the membership values of the inputs.
        '''

        super(SigmoidMF, self).__init__()

        self.n_sets = n_sets
        self.uod = uod
        
        step = ((uod[1] - uod[0]) / (n_sets - 1))
        delta = uod[1] - uod[0]
    
        self.c = nn.Parameter(torch.linspace(uod[0] + 0.05 * delta, uod[1] - 0.05 * delta, n_sets))
        self.a = nn.Parameter(2 * step * torch.ones(n_sets))

    def forward(self, x):
        x = x.reshape(-1, 1).repeat(1, self.n_sets)
        c = torch.minimum(torch.maximum(self.c, torch.tensor(self.uod[0])), torch.tensor(self.uod[1]))
        memberships = 1 / (1 + torch.exp(- self.a * (x - c)))
        return memberships


import pytest
import torch
from ANFISpy.layers import Consequents

n_vars = 2
n_sets = [2, 3]
n_rules = 6
n_samples = 11
n_classes = 5
seq_len = 7
x = torch.randn(n_samples, n_vars)
x_rec = torch.randn(n_samples, seq_len, n_vars)

def test_consequentreg_output():
    cons = Consequents(n_sets=n_sets, n_classes=1)
    out = cons(x)
    assert out.shape[0] == n_samples
    assert out.shape[1] == n_rules * 1

def test_consequentcla_output():
    cons = Consequents(n_sets=n_sets, n_classes=n_classes)
    out = cons(x)
    assert out.shape[0] == n_samples
    assert out.shape[1] == n_rules * n_classes
    
def test_consequentreg_rec_output():
    cons = Consequents(n_sets=n_sets, n_classes=1)
    out = cons(x_rec)
    assert out.shape[0] == n_samples
    assert out.shape[1] == seq_len
    assert out.shape[2] == n_rules * 1

def test_consequentcla_rec_output():
    cons = Consequents(n_sets=n_sets, n_classes=n_classes)
    out = cons(x_rec)
    assert out.shape[0] == n_samples
    assert out.shape[1] == seq_len
    assert out.shape[2] == n_rules * n_classes


import pytest
import torch
from ANFISpy.mfs import GaussianMF, BellMF, SigmoidMF, TriangularMF

n_sets = 4
uod = [-1, 1]
n_samples = 10
x = torch.randn(n_samples, 1)

@pytest.fixture(
    params=[GaussianMF, BellMF, SigmoidMF, TriangularMF]
)

def mf_class(request):
    return request.param

def test_initialization(mf_class):
    mf = mf_class(n_sets=n_sets, uod=uod)
    assert mf.n_sets == n_sets
    assert mf.uod == uod

def test_output(mf_class):
    mf = mf_class(n_sets=n_sets, uod=uod)
    y = mf(x)
    assert y.shape[0] == n_samples
    assert y.shape[1] == n_sets



from .utils import _print_rules, _plot_var, _plot_rules
from .mfs import GaussianMF, BellMF, SigmoidMF, TriangularMF
from .layers import Antecedents, Consequents, Inference, RecurrentInference
from .anfis import ANFIS, RANFIS, LSTMANFIS, GRUANFIS

__version__ = "0.6.0"



import torch
import torch.nn as nn

import itertools

'''
Notation.

N: batch size;
n: number of features;
m: number of output classes;
nj: number of fuzzy sets for feature j;
R: number of rules;
L: sequence length.
'''

class Antecedents(nn.Module):
    def __init__(self, n_sets, and_operator, mean_rule_activation=False):
        '''Calculates the antecedent values of the rules. Makes all possible combinations from the fuzzy sets defined for              
        each variable, considering rules of the form: var1 is set1 and ... and varn is setn.

        Args:
            n_sets:               list with the number of fuzzy sets associated to each variable.
            and_operator:         torch function for agregation of the membership values, modeling the AND operator.
            mean_rule_activation: bool to keep mean rule activation values.

        Returns:
            memberships:          tensor (n) with tensors (N, nj) containing the membership values of each variable.
        '''

        super(Antecedents, self).__init__()

        self.n_sets = n_sets
        self.n_rules = torch.prod(torch.tensor(n_sets))
        self.and_operator = and_operator
        self.combinations = list(itertools.product(*[range(i) for i in n_sets]))
        self.mean_rule_activation = []
        self.bool = mean_rule_activation

    def forward(self, memberships):
        N = memberships[0].size(0)
        antecedents = []
        for combination in self.combinations:
            mfs = [] 
            for var_index, set_index in enumerate(combination):
                mfs.append(memberships[var_index][:, set_index])
            weight = self.and_operator(torch.stack(mfs, dim=1), dim=1)
            if isinstance(weight, tuple):  
                weight = weight[0]  
            antecedents.append(weight)
        antecedents = torch.stack(antecedents, dim=1)
        
        if self.bool:
            with torch.no_grad():
                self.mean_rule_activation.append(torch.mean(antecedents, dim=0))    
        
        return antecedents

class Consequents(nn.Module):
    def __init__(self, n_sets, n_classes):
        '''Calculates the consequents, considering a linear combination of the input variables.

        Args:
            n_sets:       list with the number of fuzzy sets associated to each variable.
            n_classes:    int with number of classes.

        Returns:
            consequents:  tensor (N, R*m) or (N, L, R*m) containing the consequents of each rule.
        '''

        super(Consequents, self).__init__()

        self.n_vars = len(n_sets)
        self.n_rules = torch.prod(torch.tensor(n_sets))
        self.n_classes = n_classes
        self.mode = 'regression' if n_classes == 1 else 'classification'

        if self.mode == 'regression':
            self.linear = nn.Linear(
                in_features=self.n_vars,
                out_features=self.n_rules,
            )
        
        if self.mode == 'classification':
            self.linear = nn.Linear(
                in_features=self.n_vars,
                out_features=self.n_rules * self.n_classes,
            )
            
    def forward(self, x):
        consequents = self.linear(x)
        return consequents

class Inference(nn.Module):
    def __init__(self, n_classes, output_activation=nn.Identity()):
        '''Performs the Takagi-Sugeno-Kang inference.
        
        Args:
            n_classes:    int with number of classes.
            output_activation: torch function.
        
        Returns:
            y_hat: tensor (N, m) with outputs of the system.
        '''
        
        super(Inference, self).__init__()
        
        self.n_classes = n_classes
        self.mode = 'regression' if n_classes == 1 else 'classification'
        self.output_activation = output_activation

    def forward(self, antecedents, consequents):
        w = antecedents / torch.sum(antecedents, dim=1, keepdim=True)
        if self.mode == 'classification':
            n_rules = w.shape[1]
            w = w.unsqueeze(-1)
            consequents = consequents.view(-1, n_rules, self.n_classes)
        y_hat = torch.sum(w * consequents, dim=1, keepdim=True).squeeze()
        return self.output_activation(y_hat)
    
class RecurrentInference(nn.Module):
    def __init__(self, n_classes, seq_len, output_activation=nn.Identity()):
        '''Performs the Takagi-Sugeno-Kang inference.
        
        Args:
            n_classes:    int with number of classes.
            output_activation: torch function.
        
        Returns:
            y_hat: tensor (N, L, m) with outputs of the system.
        '''
        
        super(RecurrentInference, self).__init__()
        
        self.n_classes = n_classes
        self.seq_len = seq_len
        self.mode = 'regression' if n_classes == 1 else 'classification'
        self.output_activation = output_activation

    def forward(self, antecedents, consequents, h=None):
        w = antecedents / torch.sum(antecedents, dim=1, keepdim=True)
        n_rules = w.shape[1]
        if self.mode == 'regression':
            consequents = consequents.view(antecedents.shape)
        if self.mode == 'classification':
            w = w.unsqueeze(-1)
            consequents = consequents.view(-1, n_rules, self.n_classes)
        y_hat = torch.sum(w * consequents, dim=1, keepdim=True).view(-1, self.seq_len, self.n_classes)
        if h is None:
            h = torch.zeros_like(y_hat)
        y_hat = y_hat + h
        return self.output_activation(y_hat)


import pytest
import torch
import numpy as np
from ANFISpy.anfis import RANFIS

n_vars = 3
n_classes = 2
mf_names = [['L', 'M', 'H']]

variables_reg = {
    'inputs': {
        'n_sets': [3, 3, 3],
        'uod': n_vars * [(0, 1)],
        'var_names': ['a', 'b', 'c'],
        'mf_names': n_vars * mf_names,
    },
    'output': {
        'var_names': 'd',
        'n_classes': 1,
    },
}

variables_cla = {
    'inputs': {
        'n_sets': [3, 3, 3],
        'uod': n_vars * [(0, 1)],
        'var_names': ['a', 'b', 'c'],
        'mf_names': n_vars * mf_names,
    },
    'output': {
        'var_names': 'd',
        'n_classes': n_classes,
    },
}

batch_size = 5
seq_len = 7

x = torch.randn(batch_size, seq_len, n_vars, dtype=torch.float32)
y_reg = torch.randn(batch_size, seq_len, 1, dtype=torch.float32)
y_cla = torch.ones(batch_size, seq_len, dtype=torch.long)

mf_type = ['gaussian', 'bell', 'sigmoid', 'triangular']

@pytest.mark.parametrize("mf_type", mf_type)
def test_initialization_regression(mf_type):
    ranfis = RANFIS(variables_reg, mf_type, seq_len)
    assert len(ranfis.input_n_sets) == 3
    assert len(ranfis.input_uod) == 3
    assert len(ranfis.input_var_names) == 3
    assert ranfis.output_var_names == 'd'
    assert ranfis.output_n_classes == 1
    
@pytest.mark.parametrize("mf_type", mf_type)
def test_initialization_classification(mf_type):
    ranfis = RANFIS(variables_cla, mf_type, seq_len)
    assert len(ranfis.input_n_sets) == 3
    assert len(ranfis.input_uod) == 3
    assert len(ranfis.input_var_names) == 3
    assert ranfis.output_var_names == 'd'
    assert ranfis.output_n_classes == n_classes

@pytest.mark.parametrize("mf_type", mf_type)
def test_regression_forward(mf_type):
    ranfis = RANFIS(variables_reg, mf_type, seq_len)
    y_pred = ranfis(x)[0]
    assert y_pred.shape == (batch_size, seq_len, 1)

@pytest.mark.parametrize("mf_type", mf_type)    
def test_classification_forward(mf_type):
    ranfis = RANFIS(variables_cla, mf_type, seq_len)
    y_pred = ranfis(x)[0]
    assert y_pred.shape == (batch_size, seq_len, n_classes)

@pytest.mark.parametrize("mf_type", mf_type)
def test_training_regression(mf_type):
    ranfis = RANFIS(variables_reg, mf_type, seq_len)
    optimizer = torch.optim.Adam(ranfis.parameters(), lr=0.01)
    criterion = torch.nn.MSELoss()
    
    for _ in range(5):  
        optimizer.zero_grad()
        y_pred = ranfis(x)[0]
        loss = criterion(y_pred, y_reg)
        loss.backward()
        optimizer.step()
        
    for name, param in ranfis.named_parameters():
        assert param.grad is not None
        assert not torch.all(param.grad == 0)
        
@pytest.mark.parametrize("mf_type", mf_type)
def test_training_classification(mf_type):
    ranfis = RANFIS(variables_cla, mf_type, seq_len)
    optimizer = torch.optim.Adam(ranfis.parameters(), lr=0.01)
    criterion = torch.nn.CrossEntropyLoss()
    
    for _ in range(5):  
        optimizer.zero_grad()
        y_pred = ranfis(x)[0].view(-1, n_classes)
        loss = criterion(y_pred, y_cla.view(-1))
        loss.backward()
        optimizer.step()
        
    for name, param in ranfis.named_parameters():
        assert param.grad is not None
        assert not torch.all(param.grad == 0)


import pytest
import torch
from ANFISpy.layers import Antecedents

n_features = 3
n_sets = [2, 5, 7]
n_samples = 11

memberships = [
    torch.randn(n_samples, n_sets[i]) for i in range(n_features)
]

def test_antecedents_initialization():
    ant = Antecedents(n_sets=n_sets, and_operator=torch.prod)
    assert ant.n_sets == n_sets
    assert ant.n_rules == n_sets[0] * n_sets[1] * n_sets[2]
    
def test_antecedents_output():
    ant = Antecedents(n_sets=n_sets, and_operator=torch.prod, mean_rule_activation=True)
    weights = ant(memberships)
    assert weights.shape[0] == n_samples
    assert weights.shape[1] == n_sets[0] * n_sets[1] * n_sets[2]
    assert ant.mean_rule_activation[0].shape[0] == n_sets[0] * n_sets[1] * n_sets[2]



from setuptools import setup, find_packages

setup(
    name="anfispy",
    version="0.6.0",
    author="Matheus Zaia Monteiro",
    author_email="matheus.z.monteiro@gmail.com",
    url="https://github.com/mZaiam/ANFISpy",
    packages=find_packages(),
    install_requires=[
        "torch>=2.5.1",
        "numpy>=2.1.3",
        "matplotlib>=3.10.0",
        ],
    python_requires=">=3.10.8"
)



import pytest
import torch
from ANFISpy.layers import Inference, RecurrentInference

n_rules = 2
n_samples = 11
n_classes = 5
seq_len = 7

ant = torch.randn(n_samples, n_rules)
ant_rec = torch.randn(n_samples * seq_len, n_rules)
cons_reg = torch.randn(n_samples, n_rules)
cons_cla = torch.randn(n_samples, n_rules * n_classes)
cons_rec_reg = torch.randn(n_samples, seq_len, n_rules)
cons_rec_cla = torch.randn(n_samples, seq_len, n_rules * n_classes)
h_reg = torch.randn(n_samples, seq_len, 1)
h_cla = torch.randn(n_samples, seq_len, n_classes)

def test_inferencereg_output():
    inf = Inference(n_classes=1)
    out = inf(ant, cons_reg)
    assert out.shape[0] == n_samples

def test_inferencecla_output():
    inf = Inference(n_classes=n_classes)
    out = inf(ant, cons_cla)
    assert out.shape[0] == n_samples
    assert out.shape[1] == n_classes
    
def test_inferencereg_rec_output():
    inf = RecurrentInference(n_classes=1, seq_len=seq_len)
    out = inf(ant_rec, cons_rec_reg, h_reg)
    assert out.shape[0] == n_samples
    assert out.shape[1] == seq_len
    assert out.shape[2] == 1

def test_inferencecla_rec_output():
    inf = RecurrentInference(n_classes=n_classes, seq_len=seq_len)
    out = inf(ant_rec, cons_rec_cla, h_cla)
    assert out.shape[0] == n_samples
    assert out.shape[1] == seq_len
    assert out.shape[2] == n_classes


import torch
import torch.nn as nn

import matplotlib.pyplot as plt
import numpy as np

from .mfs import GaussianMF, BellMF, SigmoidMF, TriangularMF
from .layers import Antecedents, Consequents, Inference, RecurrentInference
from .utils import _plot_var, _plot_rules, _print_rules

class ANFIS(nn.Module):
    def __init__(
        self, 
        variables, 
        mf_shape, 
        and_operator=torch.prod, 
        output_activation=nn.Identity(), 
        mean_rule_activation=False
    ):
        '''Adaptative Neuro-Fuzzy Inference System with Takagi-Sugeno-Kang architecture. Can perform both regression and
           classification.

        Args:
            variables:            dict with two keys, 'inputs' and 'output'. The 'input' has a dict as its value,
                                  containing four keys: 'n_sets', 'uod', 'var_names' and 'mf_names'. They have lists as                                       
                                  their values, containing, respectively: int with number of fuzzy sets associated to the                                     
                                  variable, tuple/list with the universe of discourse of the variable, str with the name of                                   
                                  the variable and list of str with the name of the fuzzy sets. The lists need to be the                                       
                                  same length, and the index of them are all associated, that is, index 0 represents the                                       
                                  information of the same variable. Now, 'output' has only the keys 'var_names' and                                           
                                  'n_classes', with a str representing the name of the variable and an int with the number                                     
                                  of classes (if the model is a regressor, insert 1). 
            mf_shape:             str containing the shape of the fuzzy sets of the system. Supports 'triangular', 'bell'
                                  'sigmoid' and 'gaussian'.
            and_operator:         torch function to model the AND in the antecedents calculation.
            output_activation:    torch function for output activation function.
            mean_rule_activation: bool to keep the mean rule activation values. 
        '''

        super(ANFIS, self).__init__()

        self.input_n_sets = variables['inputs']['n_sets']
        self.input_uod = variables['inputs']['uod']
        self.input_var_names = variables['inputs']['var_names']
        self.input_mf_names = variables['inputs']['mf_names']
        
        self.output_var_names = variables['output']['var_names']
        self.output_n_classes = variables['output']['n_classes']
        
        self.mf_shape = mf_shape
        self.and_operator = and_operator
        
        if mf_shape == 'gaussian':
            self.memberships = nn.ModuleList(
                [GaussianMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        elif mf_shape == 'triangular':
            self.memberships = nn.ModuleList(
                [TriangularMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'bell':
            self.memberships = nn.ModuleList(
                [BellMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'sigmoid':
            self.memberships = nn.ModuleList(
                [SigmoidMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        self.antecedents = Antecedents(self.input_n_sets, and_operator, mean_rule_activation)
        self.consequents = Consequents(self.input_n_sets, self.output_n_classes)
        self.inference = Inference(self.output_n_classes, output_activation)
        
    def forward(self, x):
        memberships = [mf(x[:, i]) for i, mf in enumerate(self.memberships)]
        antecedents = self.antecedents(memberships)
        consequents = self.consequents(x)
        Y = self.inference(antecedents, consequents)
        
        return Y

    def plot_var(self, var_name, file_name=False):
        '''Plots the membership functions for a certain variable of the model.

        Args:
            var_name:  str with the name of the variable, written in the same way as given in dict variables.
            file_name: str with the name of the file to be saved, if desired.
        '''

        return _plot_var(self, var_name, file_name)

    def print_rules(self, mean_rule_activation=False):
        '''Returns a list with the rules of the model in str format.
        
        Args:
            mean_rule_activation: bool to return mean rule activation.
            
        Returns:
            rules:                list of str representing the rules of the system.
            mean:                 numpy array (R) with normalized mean rule activation.
        '''
        
        return _print_rules(self, mean_rule_activation)
    
    def plot_rules(
        self, 
        var_names, 
        n_points=1000, 
        thr=0.8, 
        levels=10, 
        cmap='viridis', 
        alpha=0.3,
        x_data=None,
        y_data=None,
        file_name=None,
    ):
        '''Plot the projection of the fuzzy rules in a two variable space.
        
        Args:
            var_names: list/tuple with the variables names.
            n_points:  int with 
            thr:       float between 0 and 1 with the threshold value of the fuzzy rules activation.
            levels:    same as matplotlib.pyplot.
            cmap:      same as matplotlib.pyplot.
            alpha:     same as matplotlib.pyplot.
            x_data:    data for scatter plot.
            y_data:    data for scatter plot.
            file_name: str with the name of the file to be saved, if desired.
        '''
        
        return _plot_rules(self, var_names, n_points, thr, levels, cmap, alpha, x_data, y_data, file_name)

class RANFIS(nn.Module):
    def __init__(
        self, 
        variables, 
        mf_shape, 
        seq_len,
        and_operator=torch.prod, 
        output_activation=nn.Identity(),
        bidirectional=False,
        mean_rule_activation=False
    ):
        '''Recurrent Neuro-Fuzzy Inference System with Takagi-Sugeno-Kang architecture. Can perform both regression and
           classification.

        Args:
            variables:            dict with two keys, 'inputs' and 'output'. The 'input' has a dict as its value,
                                  containing four keys: 'n_sets', 'uod', 'var_names' and 'mf_names'. They have lists as                                       
                                  their values, containing, respectively: int with number of fuzzy sets associated to the                                     
                                  variable, tuple/list with the universe of discourse of the variable, str with the name of                                   
                                  the variable and list of str with the name of the fuzzy sets. The lists need to be the                                       
                                  same length, and the index of them are all associated, that is, index 0 represents the                                       
                                  information of the same variable. Now, 'output' has only the keys 'var_names' and                                           
                                  'n_classes', with a str representing the name of the variable and an int with the number                                     
                                  of classes (if the model is a regressor, insert 1). 
            mf_shape:             str containing the shape of the fuzzy sets of the system. Supports 'triangular', 'bell'
                                  'sigmoid' and 'gaussian'.
            and_operator:         torch function to model the AND in the antecedents calculation.
            output_activation:    torch function for output activation function.
            mean_rule_activation: bool to keep the mean rule activation values. 
        '''

        super(RANFIS, self).__init__()

        self.input_n_sets = variables['inputs']['n_sets']
        self.input_uod = variables['inputs']['uod']
        self.input_var_names = variables['inputs']['var_names']
        self.input_mf_names = variables['inputs']['mf_names']
        
        self.output_var_names = variables['output']['var_names']
        self.output_n_classes = variables['output']['n_classes']
        
        self.mf_shape = mf_shape
        self.and_operator = and_operator
        self.seq_len = seq_len
        
        if mf_shape == 'gaussian':
            self.memberships = nn.ModuleList(
                [GaussianMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        elif mf_shape == 'triangular':
            self.memberships = nn.ModuleList(
                [TriangularMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'bell':
            self.memberships = nn.ModuleList(
                [BellMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'sigmoid':
            self.memberships = nn.ModuleList(
                [SigmoidMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        self.antecedents = Antecedents(self.input_n_sets, and_operator, mean_rule_activation)
        self.consequents = Consequents(self.input_n_sets, self.output_n_classes)
        self.inference = RecurrentInference(self.output_n_classes, self.seq_len, output_activation)
        self.recurrent = nn.RNN(
            input_size=self.antecedents.n_rules * self.output_n_classes,
            hidden_size=self.output_n_classes,
            batch_first=True,
            bidirectional=bidirectional,
        )
    
    def forward(self, x, h=None):
        memberships = [mf(x[:, :, i]) for i, mf in enumerate(self.memberships)]
        antecedents = self.antecedents(memberships)
        consequents = self.consequents(x)
        h, h_n = self.recurrent(consequents)
        y_hat = self.inference(antecedents, consequents, h)
        return y_hat, h_n
    
    def plot_var(self, var_name, file_name=False):
        '''Plots the membership functions for a certain variable of the model.

        Args:
            var_name:  str with the name of the variable, written in the same way as given in dict variables.
            file_name: str with the name of the file to be saved, if desired.
        '''

        return _plot_var(self, var_name, file_name)

    def print_rules(self, mean_rule_activation=False):
        '''Returns a list with the rules of the model in str format.
        
        Args:
            mean_rule_activation: bool to return mean rule activation.
            
        Returns:
            rules:                list of str representing the rules of the system.
            mean:                 numpy array (R) with normalized mean rule activation.
        '''
        
        return _print_rules(self, mean_rule_activation)
    
    def plot_rules(
        self, 
        var_names, 
        n_points=1000, 
        thr=0.8, 
        levels=10, 
        cmap='viridis', 
        alpha=0.3,
        x_data=None,
        y_data=None,
        file_name=None,
    ):
        '''Plot the projection of the fuzzy rules in a two variable space.
        
        Args:
            var_names: list/tuple with the variables names.
            n_points:  int with 
            thr:       float between 0 and 1 with the threshold value of the fuzzy rules activation.
            levels:    same as matplotlib.pyplot.
            cmap:      same as matplotlib.pyplot.
            alpha:     same as matplotlib.pyplot.
            x_data:    data for scatter plot.
            y_data:    data for scatter plot.
            file_name: str with the name of the file to be saved, if desired.
        '''
        
        return _plot_rules(self, var_names, n_points, thr, levels, cmap, alpha, x_data, y_data, file_name)
    
class LSTMANFIS(nn.Module):
    def __init__(
        self, 
        variables, 
        mf_shape, 
        seq_len,
        and_operator=torch.prod, 
        output_activation=nn.Identity(),
        bidirectional=False,
        mean_rule_activation=False
    ):
        '''Long-Short Term Memory Adaptative Neuro-Fuzzy Inference System with Takagi-Sugeno-Kang architecture. Can perform both regression and
           classification.

        Args:
            variables:            dict with two keys, 'inputs' and 'output'. The 'input' has a dict as its value,
                                  containing four keys: 'n_sets', 'uod', 'var_names' and 'mf_names'. They have lists as                                       
                                  their values, containing, respectively: int with number of fuzzy sets associated to the                                     
                                  variable, tuple/list with the universe of discourse of the variable, str with the name of                                   
                                  the variable and list of str with the name of the fuzzy sets. The lists need to be the                                       
                                  same length, and the index of them are all associated, that is, index 0 represents the                                       
                                  information of the same variable. Now, 'output' has only the keys 'var_names' and                                           
                                  'n_classes', with a str representing the name of the variable and an int with the number                                     
                                  of classes (if the model is a regressor, insert 1). 
            mf_shape:             str containing the shape of the fuzzy sets of the system. Supports 'triangular', 'bell'
                                  'sigmoid' and 'gaussian'.
            and_operator:         torch function to model the AND in the antecedents calculation.
            output_activation:    torch function for output activation function.
            mean_rule_activation: bool to keep the mean rule activation values. 
        '''

        super(LSTMANFIS, self).__init__()

        self.input_n_sets = variables['inputs']['n_sets']
        self.input_uod = variables['inputs']['uod']
        self.input_var_names = variables['inputs']['var_names']
        self.input_mf_names = variables['inputs']['mf_names']
        
        self.output_var_names = variables['output']['var_names']
        self.output_n_classes = variables['output']['n_classes']
        
        self.mf_shape = mf_shape
        self.and_operator = and_operator
        self.seq_len = seq_len
        
        if mf_shape == 'gaussian':
            self.memberships = nn.ModuleList(
                [GaussianMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        elif mf_shape == 'triangular':
            self.memberships = nn.ModuleList(
                [TriangularMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'bell':
            self.memberships = nn.ModuleList(
                [BellMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'sigmoid':
            self.memberships = nn.ModuleList(
                [SigmoidMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        self.antecedents = Antecedents(self.input_n_sets, and_operator, mean_rule_activation)
        self.consequents = Consequents(self.input_n_sets, self.output_n_classes)
        self.inference = RecurrentInference(self.output_n_classes, self.seq_len, output_activation)
        self.recurrent = nn.LSTM(
            input_size=self.antecedents.n_rules * self.output_n_classes,
            hidden_size=self.output_n_classes,
            batch_first=True,
            bidirectional=bidirectional,
        )
    
    def forward(self, x, hc=None):
        memberships = [mf(x[:, :, i]) for i, mf in enumerate(self.memberships)]
        antecedents = self.antecedents(memberships)
        consequents = self.consequents(x)
        h, (h_n, c_n) = self.recurrent(consequents)
        y_hat = self.inference(antecedents, consequents, h)
        return y_hat, (h_n, c_n)
    
    def plot_var(self, var_name, file_name=False):
        '''Plots the membership functions for a certain variable of the model.

        Args:
            var_name:  str with the name of the variable, written in the same way as given in dict variables.
            file_name: str with the name of the file to be saved, if desired.
        '''

        return _plot_var(self, var_name, file_name)

    def print_rules(self, mean_rule_activation=False):
        '''Returns a list with the rules of the model in str format.
        
        Args:
            mean_rule_activation: bool to return mean rule activation.
            
        Returns:
            rules:                list of str representing the rules of the system.
            mean:                 numpy array (R) with normalized mean rule activation.
        '''
        
        return _print_rules(self, mean_rule_activation)
    
    def plot_rules(
        self, 
        var_names, 
        n_points=1000, 
        thr=0.8, 
        levels=10, 
        cmap='viridis', 
        alpha=0.3,
        x_data=None,
        y_data=None,
        file_name=None,
    ):
        '''Plot the projection of the fuzzy rules in a two variable space.
        
        Args:
            var_names: list/tuple with the variables names.
            n_points:  int with 
            thr:       float between 0 and 1 with the threshold value of the fuzzy rules activation.
            levels:    same as matplotlib.pyplot.
            cmap:      same as matplotlib.pyplot.
            alpha:     same as matplotlib.pyplot.
            x_data:    data for scatter plot.
            y_data:    data for scatter plot.
            file_name: str with the name of the file to be saved, if desired.
        '''
        
        return _plot_rules(self, var_names, n_points, thr, levels, cmap, alpha, x_data, y_data, file_name)
    
class GRUANFIS(nn.Module):
    def __init__(
        self, 
        variables, 
        mf_shape, 
        seq_len,
        and_operator=torch.prod, 
        output_activation=nn.Identity(),
        bidirectional=False,
        mean_rule_activation=False
    ):
        '''Long-Short Term Memory Adaptative Neuro-Fuzzy Inference System with Takagi-Sugeno-Kang architecture. Can perform both regression and
           classification.

        Args:
            variables:            dict with two keys, 'inputs' and 'output'. The 'input' has a dict as its value,
                                  containing four keys: 'n_sets', 'uod', 'var_names' and 'mf_names'. They have lists as                                       
                                  their values, containing, respectively: int with number of fuzzy sets associated to the                                     
                                  variable, tuple/list with the universe of discourse of the variable, str with the name of                                   
                                  the variable and list of str with the name of the fuzzy sets. The lists need to be the                                       
                                  same length, and the index of them are all associated, that is, index 0 represents the                                       
                                  information of the same variable. Now, 'output' has only the keys 'var_names' and                                           
                                  'n_classes', with a str representing the name of the variable and an int with the number                                     
                                  of classes (if the model is a regressor, insert 1). 
            mf_shape:             str containing the shape of the fuzzy sets of the system. Supports 'triangular', 'bell'
                                  'sigmoid' and 'gaussian'.
            and_operator:         torch function to model the AND in the antecedents calculation.
            output_activation:    torch function for output activation function.
            mean_rule_activation: bool to keep the mean rule activation values. 
        '''

        super(GRUANFIS, self).__init__()

        self.input_n_sets = variables['inputs']['n_sets']
        self.input_uod = variables['inputs']['uod']
        self.input_var_names = variables['inputs']['var_names']
        self.input_mf_names = variables['inputs']['mf_names']
        
        self.output_var_names = variables['output']['var_names']
        self.output_n_classes = variables['output']['n_classes']
        
        self.mf_shape = mf_shape
        self.and_operator = and_operator
        self.seq_len = seq_len
        
        if mf_shape == 'gaussian':
            self.memberships = nn.ModuleList(
                [GaussianMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        elif mf_shape == 'triangular':
            self.memberships = nn.ModuleList(
                [TriangularMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'bell':
            self.memberships = nn.ModuleList(
                [BellMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        if mf_shape == 'sigmoid':
            self.memberships = nn.ModuleList(
                [SigmoidMF(n_sets_, uod_) for n_sets_, uod_ in zip(self.input_n_sets, self.input_uod)]
            )
            
        self.antecedents = Antecedents(self.input_n_sets, and_operator, mean_rule_activation)
        self.consequents = Consequents(self.input_n_sets, self.output_n_classes)
        self.inference = RecurrentInference(self.output_n_classes, self.seq_len, output_activation)
        self.recurrent = nn.GRU(
            input_size=self.antecedents.n_rules * self.output_n_classes,
            hidden_size=self.output_n_classes,
            batch_first=True,
            bidirectional=bidirectional,
        )
    
    def forward(self, x, hc=None):
        memberships = [mf(x[:, :, i]) for i, mf in enumerate(self.memberships)]
        antecedents = self.antecedents(memberships)
        consequents = self.consequents(x)
        h, h_n = self.recurrent(consequents)
        y_hat = self.inference(antecedents, consequents, h)
        return y_hat, h_n
    
    def plot_var(self, var_name, file_name=False):
        '''Plots the membership functions for a certain variable of the model.

        Args:
            var_name:  str with the name of the variable, written in the same way as given in dict variables.
            file_name: str with the name of the file to be saved, if desired.
        '''

        return _plot_var(self, var_name, file_name)

    def print_rules(self, mean_rule_activation=False):
        '''Returns a list with the rules of the model in str format.
        
        Args:
            mean_rule_activation: bool to return mean rule activation.
            
        Returns:
            rules:                list of str representing the rules of the system.
            mean:                 numpy array (R) with normalized mean rule activation.
        '''
        
        return _print_rules(self, mean_rule_activation)
    
    def plot_rules(
        self, 
        var_names, 
        n_points=1000, 
        thr=0.8, 
        levels=10, 
        cmap='viridis', 
        alpha=0.3,
        x_data=None,
        y_data=None,
        file_name=None,
    ):
        '''Plot the projection of the fuzzy rules in a two variable space.
        
        Args:
            var_names: list/tuple with the variables names.
            n_points:  int with 
            thr:       float between 0 and 1 with the threshold value of the fuzzy rules activation.
            levels:    same as matplotlib.pyplot.
            cmap:      same as matplotlib.pyplot.
            alpha:     same as matplotlib.pyplot.
            x_data:    data for scatter plot.
            y_data:    data for scatter plot.
            file_name: str with the name of the file to be saved, if desired.
        '''
        
        return _plot_rules(self, var_names, n_points, thr, levels, cmap, alpha, x_data, y_data, file_name)